// NumaMemoryMap.cpp : implementation file
//

#include "stdafx.h"
#include "resource.h"
#include "NumaMemoryMap.h"
#include <limits.h>
#include "GdiFlush.h"
#include "ToString.h"

// CNumaMemoryMap

IMPLEMENT_DYNAMIC(CNumaMemoryMap, CStatic)


/****************************************************************************
*                       CNumaMemoryMap::CNumaMemoryMap
*                       CNumaMemoryMap::~CNumaMemoryMap
* Result: 

*       
* Effect: 
*       Constructor/Destructor
****************************************************************************/

CNumaMemoryMap::CNumaMemoryMap()
   {
    SYSTEM_INFO SystemInfo;
    GetSystemInfo (&SystemInfo);
    //NumberOfProcessors = SystemInfo.dwNumberOfProcessors;
    PageSize = SystemInfo.dwPageSize;
   }

CNumaMemoryMap::~CNumaMemoryMap()
   {
   }


/****************************************************************************
*                                 Message Map
****************************************************************************/

BEGIN_MESSAGE_MAP(CNumaMemoryMap, CStatic)
    ON_WM_PAINT()
    ON_WM_SIZE()
    ON_WM_ERASEBKGND()
END_MESSAGE_MAP()

/*****************************************************************************
*                                   Node Colors
*****************************************************************************/

static const
   struct {
      COLORREF Memory;
      COLORREF Preferred;
} NodeColors[] = {
   { RGB(255,   0,   0),  RGB(255, 150, 150) },    // Red, pale red
   { RGB(  0, 255,   0),  RGB(150, 255, 150) },    // Green, pale green
   { RGB(  0,   0, 255),  RGB(150, 150, 255) },    // Blue, pale blue
   { RGB(255, 255,   0),  RGB(255, 255, 160) },    // Yellow, pale yellow

   { RGB(255,   0, 255),  RGB(255, 160, 155) },    // Magenta, pale magenta
   { RGB(  0, 255, 255),  RGB(170, 255, 255) },    // Cyan, pale cyan
   { RGB(128,   0,   0),  RGB(200,   0,   0) },    // Brown, light brown
   { RGB(  0, 128,   0),  RGB(  0, 180,   0) },    // Dark green, pale dark green

   { RGB(  0,   0, 128),  RGB(  0,   0, 230) },    // Dark blue, pale dark blue
   { RGB(128,   0, 255),  RGB(190, 128, 255) },    // Purple, pale purple
   { RGB(255, 128,   0),  RGB(255, 192, 128) },    // Orange, pale orange
   { RGB(  0, 128, 128),  RGB(128, 255, 255) }     // Dark cyan, pale dark cyan
};

/****************************************************************************
*                           CNumaMemoryMap::DrawCPU
* Inputs:
*       CDC & dc: DC to draw into
*       const CRect & rect: Rectangle in which to draw the CPU
*       const CPU & CpuInfo: CPU block
* Result: void
*       
* Effect: 
*       Draws the CPU memory map
****************************************************************************/

void CNumaMemoryMap::DrawCPU(CDC & dc, const CRect & rect, const AddressRange & range, const CPU & CpuInfo)
    {
     //================================================================
     int save = dc.SaveDC();
     UINT_PTR ScaledWidth = (range.GetHigh() - range.GetLow()) / PageSize;

     dc.FillSolidRect(&rect, NodeColors[CpuInfo.GetPreferred() % _countof(NodeColors)].Preferred);
     ::SetGraphicsMode((HDC)dc, GM_ADVANCED);

     XFORM x;
     x.eM11 = (float)rect.Width() / (float)ScaledWidth ; // horizontal scaling
     x.eM12 = 0.0;
     x.eM21 = 0;
     x.eM22 = 1.0;  // vertical scaling
     x.eDx  = 0.0f; //(float)(range.GetLow() / PageSize);
     x.eDy  = 0.0f;

     dc.SetWorldTransform(&x);

     int NodeHeight = rect.Height() / CpuInfo.GetNodeCount();

     CRect bar( INT_MAX, 0, INT_MAX, 0);

     // Note that this font is created in the context of the transformation!
     CFont textFont;
     textFont.CreatePointFont(100, _T("Arial"), &dc);

     /*****************************************************************************
     *
     * rect.left, rect.top
     * |
     * v
     * *----------------------------------------------------------------  ---
     * | CPU 0  Node 0                                                     ^
     * + - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -   | NodeHeight * GetNodeCount()
     * |        Node 1                                                     v
     * +----------------------------------------------------------------  ---
     * | CPU 1  Node 0
     * + - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
     * |        Node 1
     * +---------------------------------------------------------------*
     *                                                                 ^
     *                                                                 |
     *                                           rect.right, rect.bottom 
     *
     *****************************************************************************/

     TRACE(_T(__FUNCTION__) _T(": CpuInfo.size() = %d\n"), CpuInfo.size());

     for(SIZE_T i = 0; i < CpuInfo.size(); i++)
        { /* plot each band */
         const NumaRange numarange = CpuInfo.at(i);
         numarange.Trace();
         COLORREF color = numarange.GetValid() ? NodeColors[numarange.GetNode() % _countof(NodeColors)].Memory : RGB(0, 0, 0);
         CRect r;
         r.left =  (int) ( (numarange.GetLow() - range.GetLow()) / PageSize);
         r.right = (int) ( (numarange.GetHigh()- range.GetLow()) / PageSize);
         r.top = rect.top + numarange.GetNode() * NodeHeight;
         r.bottom = r.top + NodeHeight;
         dc.FillSolidRect(&r, color);
         bar.left = r.right = max(bar.left, r.right);
         bar.top = min(bar.top, r.top);
         bar.bottom = max(bar.bottom, r.bottom);
         GDI_FLUSH();
         // Plot a rectangle
        } /* plot each band */

     dc.MoveTo(bar.left, bar.top);
     dc.LineTo(bar.left, bar.bottom);

     dc.RestoreDC(save);
     //================================================================
     // Write the captions
     //================================================================
     save = dc.SaveDC();

     dc.SelectStockObject(DEFAULT_GUI_FONT);

     //----------------------------------------------------------------
     // If the font needs to be smaller, make it smaller
     CFont * f = dc.GetCurrentFont();
     LOGFONT lf;
     CFont newfont;

     f->GetLogFont(&lf);
     if(abs(lf.lfHeight) >= NodeHeight)
        { /* make smaller */
         lf.lfHeight = -(int)(0.9f * (double)NodeHeight);
         newfont.CreateFontIndirect(&lf);
         dc.SelectObject(&newfont);
        } /* make smaller */
     //----------------------------------------------------------------
     
     dc.SetBkMode(TRANSPARENT);
     CString cpustr = ToString(_T("CPU %d "), CpuInfo.GetCPU());
     int L = 0;
     dc.TextOut(L, rect.top, cpustr);
     GDI_FLUSH();
     CSize sz = dc.GetTextExtent(cpustr);

     for(SIZE_T i = 0; i < CpuInfo.GetNodeCount(); i++)
        { /* captions */
         if(i != 0) // && i != CpuInfo.GetNodeCount() - 1)
            { /* node sep */
             int save2 = dc.SaveDC();
             CPen pen (PS_DASH, 0, RGB(128, 128, 128));
             dc.SelectObject(&pen);
             dc.MoveTo(0, rect.top + i * NodeHeight);
             dc.LineTo(rect.right, rect.top + i * NodeHeight);
             dc.RestoreDC(save2);
            } /* node sep */
         CString nodestr = ToString(_T("Node %d"), i);
         dc.TextOut(L + sz.cx, rect.top + i * NodeHeight, nodestr);
         GDI_FLUSH();
        } /* captions */

     dc.RestoreDC(save);
     //================================================================
     // Draw the line at the bottom
     //================================================================
     save = dc.SaveDC();
     dc.MoveTo(0, rect.top);
     dc.LineTo(rect.right, rect.top);

     dc.MoveTo(0, rect.bottom);
     dc.LineTo(rect.right, rect.bottom);
     GDI_FLUSH();
     dc.RestoreDC(save);
     //================================================================
    } // CNumaMemoryMap::DrawCPU

/****************************************************************************
*                           CNumaMemoryMap::OnPaint
* Result: void
*       
* Effect: 
*       Draws the map
* Notes:
*       +---------------------------+
*      C|##   ####    ###           | Node 0
*      P|  $$$             $$$      | Node 1
*      U|         ****   **         | Node 2
*      0|                     @@@@@@| Node 3
*       +---------------------------+
*      C|         ####     ###      | Node 0
*      P|$$$$$        $$$         $$| Node 1
*      U|     ****            ****  | Node 2
*      1|                @@         | Node 3
*       +---------------------------+
****************************************************************************/

void CNumaMemoryMap::OnPaint()
    {
     CPaintDC dc(this);
     CRect client;
     GetClientRect(&client);
     if(Processors.size() == 0)
         return; // No processors, nothing to paint

     // Compute the height of a node
     int height = client.Height() / (int)Processors.size();
     for(SIZE_T i = 0; i < Processors.size(); i++)
        { /* draw each CPU */
         TRACE(_T("------------------CPU %d----------------------\n"), i);
         DrawCPU(dc,
                 CRect(client.left, client.top + (height * i), client.right, client.top + ((height) * (i + 1))),
                 Processors.GetRange(),
                 Processors.GetCPU(i));
        } /* draw each CPU */
     
    } // CNumaMemoryMap::OnPaint

/****************************************************************************
*                            CNumaMemoryMap::Clear
* Result: void
*       
* Effect: 
*       Clears the information
****************************************************************************/

void CNumaMemoryMap::Clear()
    {
     Processors.clear();
     Invalidate();
    } // CNumaMemoryMap::Clear

/****************************************************************************
*                          CNumaMemoryMap::AddRange
* Inputs:
*       UCHAR Cpu: CPU doing the allocation
*       UCHAR Nodes: Number of nodes in system
*       UINT_PTR Low: Low end of range
*       UINT_PTR High: High end of range
*       UCHAR Node: Node memory is assigned to
*       BOOL Valid: Valid range
* Result: void
*       
* Effect: 
*       Adds the element to the range
****************************************************************************/

void CNumaMemoryMap::AddRange(UCHAR Cpu, UCHAR Nodes, UINT_PTR Low, UINT_PTR High, UCHAR Node, BOOL Valid)
    {
     Processors.Add(Cpu, Nodes, Low, High, Node, Valid);
     Invalidate();
    } // CNumaMemoryMap::AddRange

/****************************************************************************
*                           CNumaMemoryMap::OnSize
* Inputs:
*       UINT nType:
*       int cx:
*       int cy:
* Result: void
*       
* Effect: 
*       Forces an invalidation
****************************************************************************/

void CNumaMemoryMap::OnSize(UINT nType, int cx, int cy)
    {
     CStatic::OnSize(nType, cx, cy);
     Invalidate();
    } // CNumaMemoryMap::OnSize

/****************************************************************************
*                        CNumaMemoryMap::OnEraseBkgnd
* Inputs:
*       CDC * pDC:
* Result: BOOL
*       TRUE, always
* Effect: 
*       Erases the background
****************************************************************************/

BOOL CNumaMemoryMap::OnEraseBkgnd(CDC * pDC)
    {
     CRect r;
     GetClientRect(&r);
     pDC->FillSolidRect(&r, ::GetSysColor(COLOR_3DFACE));
     return TRUE;
    } // CNumaMemoryMap::OnEraseBkgnd
