#pragma once

#include <vector>
#include <limits.h>
// CNumaMemoryMap

class AddressRange {
    public:
       AddressRange() { Low = SIZE_MAX; High = 0; }
       AddressRange(UINT_PTR L, UINT_PTR H) { Low = L; High = H; }
       AddressRange(const AddressRange & r) { Low = r.Low; High = r.High; }
       void Set(UINT_PTR L, UINT_PTR H) {
                                         Low = min(Low, L);
                                         High = max(High, H);
       };
       UINT_PTR GetLow() const { return Low; }
       UINT_PTR GetHigh() const { return High; }
       void Trace() const {
          TRACE(_T("%p - %p"), Low, High);
       }
    protected:
       UINT_PTR Low;
       UINT_PTR High;
};       

class CNumaMemoryMap : public CStatic
{
        DECLARE_DYNAMIC(CNumaMemoryMap)

public:
        CNumaMemoryMap();
        virtual ~CNumaMemoryMap();
        void Clear();
        void AddRange(UCHAR Cpu, UCHAR Nodes, UINT_PTR Low, UINT_PTR High, UCHAR Node, BOOL Valid);

protected:
   //****************************************************************
   //                       class NumaRange
   //****************************************************************
   class NumaRange {
       public:
          NumaRange() { Node = 0; Valid = FALSE; }
          NumaRange(UINT_PTR L, UINT_PTR H, UCHAR N, BOOL V) { Range.Set(L, H); Node = N; Valid = V; }
          UCHAR GetNode() const { return Node; }
          BOOL GetValid() const { return Valid; }
          UINT_PTR GetLow() const { return Range.GetLow(); }
          UINT_PTR GetHigh() const { return Range.GetHigh(); }
          void Trace() const {
#ifdef _DEBUG
             TRACE(_T("Node: %d: "), Node);
             Range.Trace();
             TRACE(Valid ? _T(" Valid") : _T(" Not valid"));
             TRACE(_T("\n"));
#endif
          }
       protected:
          AddressRange Range;
          UCHAR Node;
          BOOL Valid;
   };

   //****************************************************************
   //                        class CPU
   //****************************************************************
   class CPU {
       public:
          CPU() { Nodes = 0; Preferred = 0; }
          void Add(NumaRange rng, UCHAR cpu, UCHAR nodecount) {
                                   Range.Set(rng.GetLow(), rng.GetHigh());
                                   Memory.push_back(rng);
                                   ::GetNumaProcessorNode(cpu, &Preferred);
                                   Nodes = nodecount;
                                   Cpu = cpu;
                                   TRACE(_T(__FUNCTION__) _T(": "));
                                   rng.Trace();
                                   TRACE(_T(" size = %d\n"), Memory.size());
                                  }
          UINT_PTR GetLow()    const { return Range.GetLow(); }
          UINT_PTR GetHigh()   const { return Range.GetHigh(); }
          UCHAR GetNodeCount() const { return Nodes; }
          UCHAR GetCPU()       const { return Cpu; }

          SIZE_T size() const { return Memory.size(); }
          const NumaRange at(SIZE_T n) const { return Memory.at(n); }
          UCHAR GetPreferred() const { return Preferred; }
       protected:
          std::vector<NumaRange> Memory;
          AddressRange Range;
          UCHAR Nodes;
          UCHAR Preferred;
          UCHAR Cpu;
   };

   //****************************************************************
   //                        class CPUs
   //****************************************************************
   class CPUs {
       public:
          CPUs() { clear(); }
          void Add(UCHAR Cpu, UCHAR nodes, UINT_PTR L, UINT_PTR H, UCHAR N, BOOL V) {
                              Range.Set(L, H);
                              if(Cpu >= CPUData.size())
                                  CPUData.resize(Cpu + 1);
                              CPUData[Cpu].Add(NumaRange(L, H, N, V), Cpu, nodes);
                             }
          UINT_PTR GetLow()  { return Range.GetLow(); }
          UINT_PTR GetHigh() { return Range.GetHigh(); }
          AddressRange GetRange() { return AddressRange(Range); }
          void clear() { CPUData.clear(); }
          SIZE_T size() { return CPUData.size(); }
          void resize(SIZE_T n) { CPUData.resize(n); }
          const CPU & GetCPU(int n) { return CPUData[n]; }
       public:
          std::vector<CPU> CPUData;
       protected:
          AddressRange Range;
   };

   CPUs Processors;
protected:
   UINT PageSize;
   void DrawCPU(CDC & dc, const CRect & rect, const AddressRange & range, const CPU & CpuInfo);

protected:
        DECLARE_MESSAGE_MAP()
        afx_msg void OnPaint();
        afx_msg void OnSize(UINT nType, int cx, int cy);
        afx_msg BOOL OnEraseBkgnd(CDC * pDC);

};


