// SystemLogicalProcessorInformationDisplay.cpp : implementation file
//

#include "stdafx.h"
#include "SystemLogicalProcessorInformationDisplay.h"
#include "ToString.h"

// CSystemLogicalProcessorInformationDisplay

IMPLEMENT_DYNAMIC(CSystemLogicalProcessorInformationDisplay, CListBox)

CSystemLogicalProcessorInformationDisplay::CSystemLogicalProcessorInformationDisplay()
{

}

CSystemLogicalProcessorInformationDisplay::~CSystemLogicalProcessorInformationDisplay()
{
}


BEGIN_MESSAGE_MAP(CSystemLogicalProcessorInformationDisplay, CListBox)
END_MESSAGE_MAP()



// CSystemLogicalProcessorInformationDisplay message handlers




/****************************************************************************
*             CSystemLogicalProcessorInformationDisplay::DrawItem
* Inputs:
*       LPDRAWITEMSTRUCT dis:
* Result: void
*       
* Effect: 
*       Draws each element
****************************************************************************/

void CSystemLogicalProcessorInformationDisplay::DrawItem(LPDRAWITEMSTRUCT dis)
   {
    PSYSTEM_LOGICAL_PROCESSOR_INFORMATION lpi = (PSYSTEM_LOGICAL_PROCESSOR_INFORMATION)dis->itemData;

    CRect r = dis->rcItem;
    CDC * dc = CDC::FromHandle(dis->hDC);
    int save = dc->SaveDC(); 

    if(dis->itemID == (UINT)-1)
       { /* no items */
        dc->FillSolidRect(&dis->rcItem, ::GetSysColor(COLOR_WINDOW));
        if(dis->itemState & ODS_FOCUS)
           { /* selected */
            dc->DrawFocusRect(&dis->rcItem);
           } /* selected */
        return;
       } /* no items */

    COLORREF fill;
    COLORREF text;

    int gap = 2 * ::GetSystemMetrics(SM_CXBORDER);
    int offset = -GetScrollPos(SB_HORZ);

    if(dis->itemState & ODS_SELECTED)
       { /* selected */
        //if(dis->itemState & (ODS_DISABLED | ODS_GRAYED))
        if(!IsWindowEnabled())                            
           { /* selected, disabled */                     
            fill = ::GetSysColor(COLOR_GRAYTEXT);         
            text = ::GetSysColor(COLOR_WINDOWTEXT);       
           } /* selected, disabled */                     
        else                                              
           { /* selected, enabled */
            if(::GetFocus() == m_hWnd)
               { /* selected, enabled, has focus */
                fill = ::GetSysColor(COLOR_HIGHLIGHT);
                text = ::GetSysColor(COLOR_HIGHLIGHTTEXT);
               } /* selected, enabled, has focus */
            else
               { /* selected, enabled, no focus */
                fill = ::GetSysColor(COLOR_3DFACE);
                text = ::GetSysColor(COLOR_WINDOWTEXT);
               } /* selected, enabled, no focus */
           } /* selected, enabled */
       } /* selected */
    else
       { /* not selected */
        if(!IsWindowEnabled())                       
           { /* not selected, disabled */            
            fill = ::GetSysColor(COLOR_3DFACE);      
            text = ::GetSysColor(COLOR_GRAYTEXT);    
           } /* not selected, disabled */            
        else                                         
           { /* not selected, enabled */
            fill = ::GetSysColor(COLOR_WINDOW);
            text = ::GetSysColor(COLOR_WINDOWTEXT);
           } /* not selected, enabled */
       } /* not selected */

     dc->FillSolidRect(&r, fill);
     dc->SetBkMode(TRANSPARENT);
     dc->SetTextColor(text);

     if((HFONT)FixedFont == NULL)
        { /* has no font */
         CFont * f = GetFont();
         LOGFONT lf;
         f->GetLogFont(&lf);
         _tcscpy_s(lf.lfFaceName, _countof(lf.lfFaceName), _T("Courier New"));
         FixedFont.CreateFontIndirect(&lf);
        } /* has no font */

     dc->SetTextAlign(TA_RIGHT | TA_UPDATECP);

     int col1;
     CPoint pt;
     { /* processor mask */
      int save2 = dc->SaveDC();

      dc->SelectObject(&FixedFont);

      CString processor;

      if(sizeof(UINT_PTR) == sizeof(int))
         { /* 32-bit */
          col1 = dc->GetTextExtent(_T(" 00000000")).cx;

          processor.Format(_T("%08X"), lpi->ProcessorMask);
         } /* 32-bit */
      else
         { /* 64-bit */
          col1 = dc->GetTextExtent(_T(" 0000000000000000")).cx;

          processor.Format(_T("%016I64X"), lpi->ProcessorMask);
         } /* 64-bit */

      dc->MoveTo(col1, r.top);
      dc->TextOut(0, 0, processor);
      pt = dc->GetCurrentPosition();
      pt.x = col1 + dc->GetTextExtent(_T(" ")).cx;
      dc->RestoreDC(save2);
     } /* processor mask */

     int col2 = 0;
     col2 = max(col2, col1 + dc->GetTextExtent(_T("  RelationCache")).cx);
     col2 = max(col2, col1 + dc->GetTextExtent(_T("  RelationNumaNode")).cx);
     col2 = max(col2, col1 + dc->GetTextExtent(_T("  RelationProcessorCore")).cx);
     col2 = max(col2, col1 + dc->GetTextExtent(_T("  RelationProcessorPackage")).cx);
     
     dc->SetTextAlign(TA_LEFT | TA_UPDATECP);
     dc->MoveTo(pt);

     CString relation;
     switch(lpi->Relationship)
        { /* relationship */
#define CASE_RELATION(x) case x: relation = _T(#x); break
         CASE_RELATION(RelationCache);
         CASE_RELATION(RelationNumaNode);
         CASE_RELATION(RelationProcessorCore);
         CASE_RELATION(RelationProcessorPackage);
#undef CASE_RELATION
        } /* relationship */

     dc->TextOut(0, 0, relation);

     pt.x = col2;
     TEXTMETRIC tm;
     dc->GetTextMetrics(&tm);
     int h = tm.tmHeight + tm.tmInternalLeading;

     switch(lpi->Relationship)
        { /* relationship details */
         case RelationCache:
            { /* RelationCache */
             int y = r.top;
             //****************
             // Level
             //****************
             dc->MoveTo(pt.x, y);
             dc->TextOut(0, 0, ToString(_T("Level %d"), lpi->Cache.Level));

             //****************
             // Associativity
             //****************
             y += h;
             dc->MoveTo(pt.x, y);
             CString assoc;
             if(lpi->Cache.Associativity == CACHE_FULLY_ASSOCIATIVE)
                assoc = _T("CACHE_FULLY_ASSOCIATIVE");
             else
                assoc = ToString(_T("%d"), lpi->Cache.Associativity);
             
             dc->TextOut(0, 0, ToString(_T("Associativity %s"), assoc));

             //****************
             // LineSize
             //****************
             y += h;
             dc->MoveTo(pt.x, y);
             dc->TextOut(0, 0, ToString(_T("LineSize %d"), lpi->Cache.LineSize));

             //****************
             // Cache type
             //****************
             y += h;
             dc->MoveTo(pt.x, y);
             CString type;
             switch(lpi->Cache.Type)
                { /* type */
#define CASE_TYPE(x) case x: type = _T(#x); break
                 CASE_TYPE(CacheUnified);
                 CASE_TYPE(CacheInstruction);
                 CASE_TYPE(CacheData);
                 CASE_TYPE(CacheTrace);
                 default:
                    type = ToString(_T("%d"), lpi->Cache.Type);
                    break;
#undef CASE_TYPE
                } /* type */
             dc->TextOut(0, 0, ToString(_T("Type %s"), type));
            } /* RelationCache */
            break;
         case RelationNumaNode:
            { /* RelationNumaNode */
             CString s;
             dc->MoveTo(pt);
             s.Format(_T("NodeNumber %d"), lpi->NumaNode.NodeNumber);
             dc->TextOut(0, 0, s);
            } /* RelationNumaNode */
            break;
         case RelationProcessorCore:
            { /* RelationProcessorCore */
             CString s;
             dc->MoveTo(pt);
             s.Format(_T("Flags %02X"), lpi->ProcessorCore.Flags);
             dc->TextOut(0, 0, s);
            } /* RelationProcessorCore */
            break;
         case RelationProcessorPackage:
            // no additional information to display
            break;
        } /* relationship details */
     dc->RestoreDC(save);

     if(dis->itemState & ODS_FOCUS)
        dc->DrawFocusRect(&dis->rcItem);
   }


/****************************************************************************
*            CSystemLogicalProcessorInformationDisplay::MeasureItem
* Inputs:
*       LPMEASUREITEMSTRUCT mis
* Result: void
*       
* Effect: 
*       
****************************************************************************/

void CSystemLogicalProcessorInformationDisplay::MeasureItem(LPMEASUREITEMSTRUCT mis)
   {
    PSYSTEM_LOGICAL_PROCESSOR_INFORMATION lpi = (PSYSTEM_LOGICAL_PROCESSOR_INFORMATION)mis->itemData;
    int lines = 1;
    if(lpi->Relationship == RelationCache)
       lines = 4;

    CClientDC dc(this);
    dc.SelectObject(GetFont());
    TEXTMETRIC tm;
    dc.GetTextMetrics(&tm);
    int h = tm.tmHeight + tm.tmInternalLeading;

    mis->itemHeight = lines * h;
   }


/****************************************************************************
*            CSystemLogicalProcessorInformationDisplay::DeleteItem
* Inputs:
*       LPDELETEITEMSTRUCT dis
* Result: void
*       
* Effect: 
*       Deletes the object
****************************************************************************/

void CSystemLogicalProcessorInformationDisplay::DeleteItem(LPDELETEITEMSTRUCT dis)
   {
    PSYSTEM_LOGICAL_PROCESSOR_INFORMATION lpi = (PSYSTEM_LOGICAL_PROCESSOR_INFORMATION)dis->itemData;
    delete lpi;

    CListBox::DeleteItem(dis);
}
