#include "stdafx.h"
#define _93E6005A_7D4E_40cc_958A_5873BF97537F
#include "XML.h"
#include "xmlexceptions.h"
#include "XMLClass.h"
#include "ToString.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#endif

static __inline BOOL IsFirstLetter(TCHAR ch) { return _istalnum(ch) || ch == _T('_'); }
static __inline BOOL IsInternalLetter(TCHAR ch) { return IsFirstLetter(ch) || ch == _T('-'); }
static __inline CString IndentOutput(int depth) { return CString(_T(' '), 3 * depth); }


#if defined(_DEBUG) && defined(TRACE_LEXER)
#define Indent(n) CString(_T(' '), 2 * (n))
#define LEX() { lexer(lexeme, token, FALSE, offset, source, line); \
              TRACE(_T("%s") _T(__FUNCTION__) _T("(%d): input (%d): %s=\"%s\"\n"), Indent(depth), __LINE__, line, cvToken(token), lexeme); }
#define LEXCOMMENT(){ lexer(lexeme, token, TRUE, offset, source, line); \
                    TRACE(_T("%s") _T(__FUNCTION__) _T("(%d): input (%d): %s=\"%s\"\n"), Indent(depth), __LINE__, line, cvToken(token), lexeme); }
#else
#define Indent(n) CString(_T(' '), 2 * (n))
#define LEX() lexer(lexeme, token, FALSE, offset, source, line)
#define LEXCOMMENT() lexer(lexeme, token, TRUE, offset, source, line)
#endif

CWnd * XML::wnd = NULL;
UINT XML::msg = 0;

/****************************************************************************
*                                 XMLNode::cvStringToXML
* Inputs:
*       const CString & text:
* Result: CString
*       XML-compatible string
* Effect: 
*       <   &lt;
*       >   &gt;
*       "   &quot;
*       &   &amp;
****************************************************************************/

/* static */ CString XMLNode::cvStringToXML(const CString & text)
    {
     CString result = text;
     result.Replace(_T("&"), _T("&amp;")); // THIS MUST BE FIRST!!!!
     
     result.Replace(_T("\xA0"), _T("&nbsp;"));
     result.Replace(_T("<"), _T("&lt;"));
     result.Replace(_T(">"), _T("&gt;"));
     result.Replace(_T("\""), _T("&quot;"));
     return result;
    } // cvStringToXML

/****************************************************************************
*                                cvXMLToString
* Inputs:
*       const CString & text: Input text in XML format
* Result: CString
*       Down-converted
* Effect: 
*       &amp;   &
*       &lt;    <
*       &gt;    >
*       &quot;  "
* Notes:
*       Maybe this should be moved to the lexer...
****************************************************************************/

CString cvXMLToString(const CString & text)
    {
     CString result = text;
     result.Replace(_T("&quot;"), _T("\""));
     result.Replace(_T("&amp;"), _T("&"));
     result.Replace(_T("&lt;"), _T("<"));
     result.Replace(_T("&gt;"), _T(">"));
     result.Replace(_T("&nbsp;"), _T("\xA0"));  // 160
     result.Replace(_T("&iexcl;"), _T(""));    // 161
     result.Replace(_T("&cent;"), _T(""));     // 162
     result.Replace(_T("&pound;"), _T(""));    // 163
     result.Replace(_T("&curren;"), _T(""));   // 164
     result.Replace(_T("&yen;"), _T(""));      // 165
     result.Replace(_T("&brvbar;"), _T(""));   // 166
     result.Replace(_T("&sect;"), _T(""));     // 167
     result.Replace(_T("&uml;"), _T(""));      // 168
     result.Replace(_T("&copy;"), _T(""));     // 169
     result.Replace(_T("&ordf;"), _T(""));     // 170
     result.Replace(_T("&laquo;"), _T(""));    // 171
     result.Replace(_T("&not;"), _T(""));      // 172
     result.Replace(_T("&shy;"), _T(""));      // 173 non-breaking hyphen

     return result;
    } // cvXMLToString

/****************************************************************************
*                                   cvText
* Inputs:
*       const CString & s:
* Result: CString
*       Replaces internal characters with \n, \r, etc.
****************************************************************************/

static CString cvText(const CString & s)
    {
     CString result = s;
     result.Replace(_T("\\"), _T("\\\\"));
     result.Replace(_T("\f"), _T("\\f"));
     result.Replace(_T("\n"), _T("\\n"));
     result.Replace(_T("\r"), _T("\\r"));
     result.Replace(_T("\t"), _T("\\t"));
     // TODO: everything else
     return result;
    } // cvText

/****************************************************************************
*                                   cvToken
* Inputs:
*       TokenType token:
* Result: CString
*       
****************************************************************************/

#ifdef _DEBUG
/* static */ CString XML::cvToken(XML::TokenType token)
    {
     switch(token)
        { /* token */
#define TokenCase(x) case x: return _T(#x)
         TokenCase(Error);
         TokenCase(OpenQuest);
         TokenCase(CloseQuest);
         TokenCase(Open);
         TokenCase(Close);
         TokenCase(OpenComment);
         TokenCase(CloseComment);
         TokenCase(SlashClose);
         TokenCase(Colon);
         TokenCase(Equal);
         TokenCase(Slash);
         TokenCase(Name);
         TokenCase(Value);
         TokenCase(Integer);
         TokenCase(Other);
         TokenCase(Empty);
#undef TokenCase
        } /* token */
     return ToString(_T("%d"), token);
    } // cvToken
#endif

/****************************************************************************
*                              XMLNode::AddChild
* Inputs:
*       XMLNode * node: Node to add
* Result: void
*       
* Effect: 
*       Adds a child node to the specified XML node
****************************************************************************/

void XMLNode::AddChild(XMLNode * node)
    {
     Children.push_back(node);
    } // XMLNode::AddChild

/****************************************************************************
*                                  XML::lexer
* Inputs:
*       CString & result: Result of parser
*       TokenType & token: Token name
*       BOOL incomment: TRUE if we are parsing a comment; FALSE otherwise
*       int & offset: Current offset into string
*       const CString & source: Source string
*       int & line: Line counter
* Result: void
*       
* Effect: 
*       Removes a token from the source string, returns it, sets its type
****************************************************************************/

/* static */ void XML::lexer(CString & lexeme, TokenType & token, BOOL incomment, int & offset, const CString & source, int & line)
   {
    lexeme = _T("");
    token = Error;
#ifdef _DEBUG
    TCHAR ch = source[offset];
#endif
    for(; offset < source.GetLength(); offset++)
       { /* scan whitespace */
        if(!_istspace(source[offset]))
           break;
        if(source[offset] == _T('\n'))
           { /* increment line */
            line++;
            if(wnd != NULL)
               { /* notification */
                wnd->SendMessage(msg, (WPARAM)offset);
               } /* notification */
           } /* increment line */
       } /* scan whitespace */
    
    //****************
    // empty
    //****************
    if(offset >= source.GetLength())
       { /* empty */
        token = Empty;
        return;
       } /* empty */
#ifdef _DEBUG
    ch = source[offset];
#endif
    //****************
    // <
    // <?
    // <!--
    //****************
    if(source[offset] == _T('<'))
       { /* open angle */
        if(offset + 1 < source.GetLength() && source[offset + 1] == _T('?'))
           { /* <? */
            token = OpenQuest;
            lexeme = _T("<?");
            offset += 2;
            return;
           } /* <? */

        if(offset + 3 < source.GetLength() && source[offset + 1] == _T('!') && source[offset + 2] == _T('-') && source[offset + 3] == _T('-'))
           { /* open comment */
            token = OpenComment;
            lexeme = _T("<!--");
            offset += 4;
            return;
           } /* open comment */

        token = Open;
        lexeme = _T("<");
        offset++;
        return;
       } /* open angle */

    //****************
    // /
    // />
    //****************
    if(source[offset] == _T('/'))
       { /* slash */
        if(offset + 1 < source.GetLength() && source[offset + 1] == _T('>'))
           { /* /> */
            token = SlashClose;
            lexeme = _T("/>");
            offset += 2;
            return;
           } /* /> */
        token = Slash;
        lexeme = _T("/");
        offset++;
        return;
       } /* slash */

    //****************
    // ?>
    //****************
    if(offset + 1 < source.GetLength() && source[offset] == _T('?') && source[offset + 1] == _T('>'))
       { /* ?> */
        token = CloseQuest;
        lexeme = _T("?>");
        offset += 2;
        return;
       } /* ?> */

    //****************
    // -->
    //****************
    if(offset + 2 < source.GetLength() &&
       source[offset] == _T('-') &&
       source[offset + 1] == _T('-') &&
       source[offset + 2] == _T('>'))
       { /* close comment */
        token = CloseComment;
        lexeme = _T("-->");
        offset += 3;
        return;
       } /* close comment */
    //****************
    // >
    //****************
    if(source[offset] == _T('>'))
       { /* colon */
        lexeme = _T(">");
        token = Close;
        offset++;
        return;
       } /* colon */

    //****************
    // :
    //****************
    if(source[offset] == _T(':'))
       { /* colon */
        lexeme = _T(":");
        token = Colon;
        offset++;
        return;
       } /* colon */

    //****************
    // =
    //****************
    if(source[offset] == _T('='))
       { /* equal */
        lexeme = _T("=");
        token = Equal;
        offset++;
        return;
       } /* equal */
    
    //****************
    // "value"
    // 'value'
    //****************
    if(!incomment && (source[offset] == _T('\'') ||
       source[offset] == _T('\"')))
       { /* quote */
        TCHAR closequote = source[offset];
        int start = ++offset;
        for(; offset < source.GetLength(); offset++)
           { /* parse string */
            if(source[offset] == closequote)
               { /* closed */
                // 1..i
                token = Value;
                lexeme = source.Mid(start, (offset - start));
                DWORD v;                         
                if(IsInteger(lexeme, v))         
                   token = Integer;  
                offset++;
                return;
               } /* closed */
           } /* parse string */

        // fell off end
        token = Error; // unclosed quoted value
        lexeme = source.Mid(start, (offset - start) - 1);
        return;
       } /* quote */
    
    //****************
    // [A-Za-z0-9_]+[-A-Za-z0-9]*
    //****************
    if(IsFirstLetter(source[offset]))
       { /* starts with letter */
        CString name;
        int start = offset;
        for(; offset < source.GetLength(); offset++)
           { /* scan name */
            if(IsInternalLetter(source[offset]))
               { /* letter */
                continue;
               } /* letter */
            lexeme = source.Mid(start, (offset - start));
            token = Name;
            return;
           } /* scan name */
       } /* starts with letter */
    //****************
    // Anything else
    //****************
    lexeme = source.Mid(offset, 1);
    offset++;
    token = Other;
    return;
    // If we get here, we have not gotten to any path that modified Token, which is Error
   } // lexer

/****************************************************************************
*                              XMLNode::GetValue
* Inputs:
*       const CString & nm: Name of value to find
* Result: CString
*       Empty string if value is not defined, or is defined and is the empty string
* Notes: 
*       To see if the value exists, see IsValue.  If
*       IsValue(_T("name")) && GetValue(_T("name")).IsEmpty()(
*       then the value is defined, but has the empty string e.g.,
*               name = ""
*       But if name is not defined at all, IsValue(_T("name")) returns FALSE
****************************************************************************/

CString XMLNode::GetValue(const CString & nm) const
    {
     XMLValue::const_iterator iter;
     iter = Values.find(nm);
         if(iter == Values.end())
            return _T("");
         return iter->second->GetValue();
    } // XMLNode::GetValue

/****************************************************************************
*                              XMLNode::SetValue
* Inputs:
*       const CString & nm: Name of value to set
*       CString & val: Value to assign to it
*       int ln = 0: Line number in original source file
* Result: BOOL
*       TRUE if the value did not previously exist, and was created
*       FALSE if the value previously existed, and was changed
* Effect: 
*       Creates a new value entry if one does not already exist, otherwise
*       replaces the existing value.  Note this is different from AddValue,
*       which throws an exception if there is a duplicate value
****************************************************************************/

BOOL XMLNode::SetValue(const CString & nm, const CString & val, int ln /* = 0 */)
    {
     XMLValue::iterator iter;
     iter = Values.find(nm);
     if(iter == Values.end())
        { /* add it */
         if(val.IsEmpty())
            return FALSE; // do not add empty string
         AddValue(nm, val, ln);
         return FALSE;
        } /* add it */
     if(val.IsEmpty())
        { /* remove it */
         Values.erase(iter);
        } /* remove it */
     else
        { /* set it */
         iter->second->SetValue(val);
        } /* set it */
     return TRUE;
    } // XMLNode::SetValue

/****************************************************************************
*                              XMLNode::AddValue
* Inputs:
*       const CString & nm: Name of value to set
*       const CString & val: Value to set
*       int ln: Line number in original source file
* Result: void
*       
* Effect: 
*       Adds a value
* Throws:
*       CXMLDuplicateValueException if the value already exists.  See
*       SetValue to allow changing the value
****************************************************************************/

void XMLNode::AddValue(const CString & nm, const CString & val, int ln /*= 0*/)
    {
     XMLValue::iterator iter;
     iter = Values.find(nm);
     if(iter == Values.end())
        { /* not found */
         Values.insert(std::pair<CString, Value *>(nm, new Value(val, ln)));
        } /* not found */
     else
        { /* found */
         throw new CXMLDuplicateValueException(iter->second->GetValue(), iter->second->GetLine(), nm, val, ln);
        } /* found */
    } // XMLNode::AddValue

/****************************************************************************
*                               XMLNode::IsTrue
* Inputs:
*       const CString & val: Value
* Result: BOOL
*       TRUE if the value is interpretable as being "True"
* Notes: 
*       y, y
*       1
*       YES, yes, etc.
*       TRUE, true, etc.
****************************************************************************/

/* static */ BOOL XMLNode::IsTrue(const CString & val) 
    {
    CString t = val;
     t.Trim();
     t.MakeUpper();
     if(t == _T("TRUE"))
        return TRUE;
     if(t == _T("1"))
        return TRUE;
     if(t == _T("Y"))
        return TRUE;
     if(t == _T("YES"))
        return TRUE;
     return FALSE;
    } // XMLNode::IsTrue

/****************************************************************************
*                              XMLNode::IsValue
* Inputs:
*       const CString & nm: Name of attribute
* Result: BOOL
*       TRUE if the attribute exists, even if it has empty value
*       FALSE if the attribute does not exist
****************************************************************************/

BOOL XMLNode::IsValue(const CString & nm) const
    {
     XMLValue::const_iterator iter;
     iter = Values.find(nm);
     return iter != Values.end();
    } // XMLNode::IsValue

/****************************************************************************
*                                XML::IsInteger
* Inputs:
*       const CString & s: String to parse
*       DWORD & value: Value converted
* Result: BOOL
*       TRUE if a number (value is the binary representation)
*       FALSE if not a number (value is undefined)
* Effect: 
*       Parses the integer
****************************************************************************/

/* static */ BOOL XML::IsInteger(const CString & s, DWORD & value)  // REQ #012
    {                                                               // REQ #012
     CString t(s);
     t.Trim();                                                      // REQ #012
     LPTSTR p = NULL;                                               // REQ #012
     value = _tcstoul(t, &p, 0);                                    // REQ #012
     if(p == NULL || *p == _T('\0'))                                // REQ #012
        return TRUE;                                                // REQ #012
     return FALSE;                                                  // REQ #012
    } // XML::IsInteger                                             // REQ #012

/****************************************************************************
*                                 XML::Build
* Inputs:
*       const CString & text: Input text
* Result: XMLNode *
*       The root node
* Effect: 
*       Parses the input string
****************************************************************************/

XMLNode * XML::Build(const CString & source)
    {
     XMLNode * node = new XMLNode;
     int offset = 0;
     line = 1;
     
     try {
        BuildNode(node, source, offset, 0);
     }
     catch(CXMLSyntaxException * e)
        { /* syntax exception */
         CString s = e->GetString();
         TRACE(_T(__FUNCTION__) _T("%s\n"), s);
         s += _T("\n");
         s += e->GetText();
         AfxMessageBox(s, MB_ICONERROR);
         e->Delete();
         delete node;
         return NULL;
        } /* syntax exception */
     catch(CXMLException * e)
        { /* failed */
         CString s = e->GetString();
         TRACE(_T(__FUNCTION__) _T("%s\n"), s);
         AfxMessageBox(s, MB_ICONERROR);
         e->Delete();
         delete node;
         return NULL;
        } /* failed */

     return node;
     
    } // XML::Build

/****************************************************************************
*                               XML::ParseClose
* Inputs:
*       XMLNode * parent:
*       const CString & source:
*       int & offset:
*       int depth: Depth, for debugging printout
* Result: void
*       
* Effect: 
*       Parses </name> or </tag:name>
****************************************************************************/

void XML::ParseClose(XMLNode * parent, const CString & source, int & offset, int depth)
   {
#ifdef TRACE_PARSER
    TRACE(_T("%s") _T(__FUNCTION__) _T("(%d): parent %p(%s:%s), \"%s\"\n"),
          Indent(depth), __LINE__, parent, ((XMLTag*)parent)->GetTag(), ((XMLTag*)parent)->GetName(), cvText(source.Mid(offset, 24)));
#endif
    CString lexeme;
    TokenType token;
    CString tag;
    CString name;
    // < /     tag : name>
    // < /     name>
    //       ^
    //       | we are here
    LEX();
    // < /     tag : name>
    // < /    name >
    //            ^
    //            | we are here                
    switch(token)
       { /* /name */
        case Name:
           tag = lexeme;
           break;
        default:
           throw new CXMLExpectedNameException(lexeme, line, source, offset);
       } /* /name */

    // < /     tag : name>
    // < /    name >
    //            ^
    //            | we are here                
    LEX();
    // < /     tag : name>
    // < /    name >
    //              ^
    //              | we are here                

    switch(token)
       { /* end colon? */
        case Colon:
           // < /     tag : name>
           //              ^
           //              | we are here                
           LEX();
           // < /     tag : name>
           //                   ^
           //                   | we are here                
           switch(token)
              { /* colon */
               case Name:
                  name = lexeme;
                  // < /     tag : name >
                  //                   ^
                  //                   | we are here                
                  LEX();
                  // < /     tag : name>
                  //                    ^
                  //                    | we are here                
                  break;
               default:
                  throw new CXMLExpectedNameException(lexeme, line, source, offset);
              } /* colon */
           break;
        case Close:
           // < /    name >
           //              ^
           //              | we are here                
           name = tag;
           tag.Empty();
           break;
        default:
           throw new CXMLExpectedTokenException(_T(":>"), lexeme, line, source, offset);
           break;
       } /* end colon? */

    // < /     tag : name>
    //                    ^
    //                    | we are here
    // < /    name >
    //              ^
    //              | we are here                
    switch(token)
       { /* close> */
        case Close:
           break;
        default:
           throw new CXMLExpectedTokenException(_T(">"), lexeme, line, source, offset);
       } /* close> */

    XMLTag * opener = dynamic_cast<XMLTag*>(parent);
    if(opener == NULL)
       { /* not in tag */
        throw new CXMLUnmatchedCloseException(_T(""), _T(""), parent->GetLine(), tag, name, line, source, offset);
       } /* not in tag */
    
    if(opener->GetTag() == tag &&
       opener->GetName() == name)
       { /* valid </name> */
        return;
       } /* valid </name> */
    else
       { /* unmatched close */
        throw new CXMLUnmatchedCloseException(opener->GetTag(), opener->GetName(), opener->GetLine(),
                                              tag, name, line, source, offset);
       } /* unmatched close */
    } // XML::ParseClose

/****************************************************************************
*                               XML::ParseNode
* Inputs:
*       XMLNode * parent:
*       const CString & source:
*       int & offset:
*       int depth: Depth, for debugging printout
* Result: BOOL
*       TRUE if the operation was a close-node, FALSE otherwise
* Effect: 
*       Creates a node and attaches it to the parent
****************************************************************************/

BOOL XML::ParseNode(XMLNode * parent, const CString & source, int & offset, int depth)
    {
     TokenType token;
     CString lexeme;

     CString tag;
     CString name;
     
     int startline = line;
     
     // < tag  : name values />
     // < name        values />
     // < tag  : name values> ... </tag : name>
     // < name        values> ... </name>
     // < / tag : name>
     // < / name>
     //  ^
     //  | we are here
     LEX();
     // < tag  : name values />
     // < name        values />
     // < tag  : name values> ... </tag : name>
     // < name        values> ... </name>
     // < /     tag : name>
     // < /     name>
     //       ^
     //       | we are here
     switch(token)
        { /* token1 */
         case Name:
            break;
         case Slash:
            ParseClose(parent, source, offset, depth);
            return TRUE;
         default:
            throw new CXMLExpectedNameException(lexeme, line, source, offset);
        } /* token1 */
     tag = lexeme;

     int back = offset;
     LEX();
     switch(token)
        { /* token */
         case Colon:
            // < tag         :   name values />
            // < tag         :   name values> ... </tag : name>
            //                  ^
            //                  | we are here
            LEX();
            // < tag         :   name values />
            // < tag         :   name values> ... </tag : name>
            //                       ^
            //                       | we are here

            switch(token)
               { /* name after colon */
                case Name:
                   name = lexeme;
                   break;
                default:
                   throw new CXMLExpectedNameException(lexeme, line, source, offset);
               } /* name after colon */
            break;
         case Name:
            // < name        name = value />
            // < name        name = value> ... </name>
            //                   ^
            //                   | we are here
            name = tag;
            tag.Empty();
            offset = back;
            break;
        } /* token */

     // < tag  : name values />
     // < name        values />
     // < tag  : name values> ... </tag : name>
     // < name        values> ... </name>
     //               ^
     //               | we are here
     XMLTag * current = new XMLTag(tag, name, startline);
     
     try { 
        ParseValues(current, source, offset, token, lexeme, depth + 1);
        switch(token)
           { /* token */
            case Close:
               BuildNode(current, source, offset, depth + 1);
               break;
            case SlashClose:
               break;
            default:
               throw new CXMLExpectedTokenException(_T("?>"), lexeme, line, source, offset);
           } /* token */
     } // try
     catch(CXMLException *)
        {
         delete current;
         throw;
        } // catch
     parent->AddChild(current);
     return FALSE;
    } // XML::ParseNode

/****************************************************************************
*                              XML::ParseValues
* Inputs:
*       XMLNode * current: Current node
*       const CString & source: XML source string
*       int & offset: Current offset into the source
*       TokenType & token: Place to put token found
*       CString & lexeme: Place to put lexeme
*       int depth: Indentation depth for debug printout
* Result: void
*       
* Effect: 
*       Parses a value list
* Syntax:
*       {name = "value"}*
****************************************************************************/

void XML::ParseValues(XMLNode * current, const CString & source, int & offset, TokenType & token, CString & lexeme, int depth)
    {
     while(TRUE)
        { /* parse value list */
         CString name;
         //   name = "value" ... >
         //   ^                  ?>
         //   |                  />
         //   we are here
         //                     ^
         //                     |
         //                     or here
         LEX();
         switch(token)
            { /* token */
             case Name:
                break;
             case CloseQuest:
             case Close:
                 return;
             case SlashClose:
                return;
             default:
                throw new CXMLExpectedNameException(lexeme, line, source, offset);
            } /* token */
         name = lexeme;
         name.Trim();

         //   name = "value" ... >
         //        ^             ?>
         //        |             />
         //        we are here
         LEX();
         switch(token)
            { /* equal? */
             case Equal:
                break;
             default:
                throw new CXMLExpectedTokenException(_T("="), lexeme, line, source, offset);
            } /* equal? */

         //   name = "value" ... >
         //         ^            ?>
         //         |            />
         //         we are here
         LEX();

         switch(token)
            { /* value? */
             case Value:
             case Integer:
                lexeme.Trim();
                lexeme = cvXMLToString(lexeme);
                if(!lexeme.IsEmpty())
                   current->AddValue(name, lexeme, line);
                break;
             default:
                 throw new CXMLExpectedValueException(lexeme, line, source, offset);
            } /* value */   
        } /* parse value list */
    } // XML::ParseValues

/****************************************************************************
*                               XML::ParseXMLHeader
* Inputs:
*       XMLNode * parent: Parent node
*       const CString & source: Source string
*       int & offset: Offset into source string
*       int depth: Depth of nesting
* Result: void
*       
* Effect: 
*       Parses an XML header
****************************************************************************/

void XML::ParseXMLHeader(XMLNode * parent, const CString & source, int & offset, int depth)
    {
     CString lexeme;
     TokenType token;

     // <?whatever values?>
     //   ^
     //   |
     //   we are here

     LEX();
     // <?whatever values?>
     //           ^
     //           |
     //           we are here

     switch(token)
        { /* token1 */
         case Name:
            break;
         default:
            throw new CXMLExpectedNameException(lexeme, line, source, offset);
        } /* token1 */

     XMLNode * current = new XMLHeader(line);

     try {
        ParseValues(current, source, offset, token, lexeme, depth + 1);
        if(token != CloseQuest)
           { /* wrong terminator */
            throw new CXMLExpectedTokenException(_T("?>"), lexeme, line, source, offset);
           } /* wrong terminator */
     }
     catch(CXMLException *)
        {
         delete current;
         throw;
        }
     parent->AddChild(current);
    } // XML::ParseXMLHeader

/****************************************************************************
*                              XML::ParseComment
* Inputs:
*       XMLNode * parent:
*       const CString & source:
*       int & offset
*       int depth:
* Result: void
*       
* Effect: 
*       Parses a comment
****************************************************************************/

void XML::ParseComment(XMLNode * parent, const CString & source, int & offset, int depth)
    {
     CString lexeme;
     TokenType token;
     int start = offset;
     int startline = line;

     while(TRUE)
        { /* scan to eoc */
         int end = offset;
         LEXCOMMENT();
         switch(token)
            { /* token */
             case CloseComment:
                {
                 CString text;
                 text = source.Mid(start, end - start);
                 text.Trim();
                 parent->AddChild(new XMLComment(text));
                 return;
                }
             case Empty:
                throw new CXMLUnterminatedCommentException(startline, source, offset);
            } /* token */
        } /* scan to eoc */
    } // XML::ParseComment

/****************************************************************************
*                               XML::BuildNode
* Inputs:
*       XMLNode * parent:
*       const CString & source:
*       int & offset:
*       int depth: Indent depth
* Result: void
*       
* Effect: 
*       Parses the node
****************************************************************************/

void XML::BuildNode(XMLNode * parent, const CString & source, int & offset, int depth)
    {
     CString lexeme;

     TokenType token;

     while(true)
        { /* scan file */
         LEX();
         if(token == Empty)
            break;
         switch(token)
            { /* token */
             case OpenQuest:
                ParseXMLHeader(parent, source, offset, depth + 1);
                break;
             case OpenComment:
                ParseComment(parent, source, offset, depth + 1);
                break;
             case Open:
                if(ParseNode(parent, source, offset, depth + 1))
                    return;
                break;
             default:
                throw new CXMLExpectedTokenException(_T("<"), lexeme, line, source, offset);
            } /* token */
        } /* scan file */
    } // XML::BuildNode

/****************************************************************************
*                                XML::UnBuild
* Inputs:
*       const XMLNode * node: Node to display
* Result: CString
*       
* Effect: 
*       
****************************************************************************/

CString XML::UnBuild(const XMLNode * node, BOOL CRLF)
    {
     CString result;
     node->UnParse(result, CRLF, 0);
     return result;
    } // XML::UnBuild

/****************************************************************************
*                               XMLTag::UnParse
* Inputs:
*       
* Result: void
*       
* Effect: 
*       <tag:name key = "value"... />   || no children
*       <tag:name key = "value"...>     || has children
*            child nodes
*       </tag:name>
****************************************************************************/

void XMLTag::UnParse(CString & result, BOOL CRLF, int depth) const
    {
     result += IndentOutput(depth);
     result += _T("<");
     if(!Tag.IsEmpty())
        { /* tag: */
         result += Tag;
         result += _T(":");
        } /* tag: */
     result += Name;

     if(!Values.empty())
        { /* has values */
         UnParseValues(result, CRLF, depth + 1);
        } /* has values */

     if(Children.empty())
        { /* no children */
         result += _T("/>");
         if(CRLF)
            result += _T("\r\n");
         else
            result += _T("\n");
        } /* no children */
     else
        { /* has children */
         result += _T(">");
         if(CRLF)
            result += _T("\r\n");
         else
            result += _T("\n");
         UnParseChildren(result, CRLF, depth + 1);
         result += IndentOutput(depth);
         result += _T("</");
         if(!Tag.IsEmpty())
            { /* show close tag */
             result += Tag;
             result += _T(":");
            } /* show close tag */
         result += Name;
         result += _T(">");
         if(CRLF)
            result += _T("\r\n");
         else
            result += _T("\n");
        } /* has children */
    } // XMLTag::UnParse

/****************************************************************************
*                             XMLComment::UnParse
* Inputs:
*       CString & result: The string being formed
*       BOOL CRLF: TRUE to use CRLF, FALSE to use LF
* Result: void
*       
* Effect: 
*       Appends the comment to the input string
****************************************************************************/

void XMLComment::UnParse(CString & result, BOOL CRLF, int depth) const
    {
     result += IndentOutput(depth);
     result += _T("<!-- ");
     result += cvStringToXML(text);
     result += _T(" -->");
     if(CRLF)
        result += _T("\r\n");
     else
        result += _T("\n");
    } // XMLComment::UnParse

/****************************************************************************
*                             XMLHeader::UnParse
* Inputs:
*       
* Result: void
*       
* Effect: 
*       
****************************************************************************/

void XMLHeader::UnParse(CString & result, BOOL CRLF, int depth) const
    {
     result += _T("<?xml-stylesheet");
     UnParseValues(result, CRLF, depth);
     result += _T("?>");
     if(CRLF)
        result += _T("\r\n");
     else
        result += _T("\n");
    } // XMLHeader::UnParse

/****************************************************************************
*                              XMLRoot::UnParse
* Inputs:
*       
* Result: void
*       
* Effect: 
*       
****************************************************************************/

void XMLRoot::UnParse(CString & result, BOOL CRLF, int depth) const
    {
     XMLNode::UnParse(result, CRLF, depth + 1);
    } // XMLRoot::UnParse

/****************************************************************************
*                              XMLNode::UnParseValues
* Inputs:
*       
* Result: 
*       
* Effect: 
*       Unparses only the values and children
****************************************************************************/

void XMLNode::UnParseValues(CString & result, BOOL CRLF, int depth) const
    {
     for(XMLValue::const_iterator iter = Values.begin(); iter != Values.end(); ++iter)
        { /* scan values */
         // TODO: put out on multiple lines
         if(iter->second->IsEmpty())
            continue; // skip empty values
         result += _T(" ");
         result += iter->first;
         result += _T(" = \"");
         result += cvStringToXML(iter->second->GetValue());
         result += _T("\"");
        } /* scan values */
    } // XMLNode::UnParseValues

/****************************************************************************
*                          XMLNode::UnParseChildren
* Inputs:
*       CString & result:
*       BOOL CRLF: TRUE to use CRLF, FALSE to use LF
*       int depth: Indentation depth
* Result: void
*       
* Effect: 
*       Appends unparse of children to result
****************************************************************************/

void XMLNode::UnParseChildren(CString & result, BOOL CRLF, int depth) const
    {
     if(Children.empty())
        { /* no children */
         result += _T("/>");
         if(CRLF)
            result += _T("\r\n");
         else
            result += _T("\n");
        } /* no children */
     else
        { /* has children */
         for(ChildNodes::const_iterator iter = Children.begin(); iter != Children.end(); ++iter)
            { /* show children */
             (*iter)->UnParse(result, CRLF, depth + 1);
            } /* show children */
        } /* has children */
    } // XMLNode::UnParseChildren

/****************************************************************************
*                                XMLNode::UnParse
* Inputs:
*       CString & result:
*       BOOL CRLF: TRUE to use CRLF, FALSE to use LF
*       int depth:
* Result: void
*       
* Effect: 
*       
****************************************************************************/

void XMLNode::UnParse(CString & result, BOOL CRLF, int depth) const
    {
     UnParseChildren(result, CRLF, depth);
    } // XMLNode::UnParse

/****************************************************************************
*                             XMLNode::CopyValues
* Inputs:
*       XMLNode * node: Input node whoae values are to be copied to this node
* Result: void
*       
* Effect: 
*       Copies all the Values array to the Values array of this node.
* Notes:
*       This REPLACES all the Values in the array; if any were already
*       present, they are unceremoniously erased and the Value nodes deleted
****************************************************************************/

void XMLNode::CopyValues(const XMLNode * node) 
    {
     for(XMLValue::iterator iter = Values.begin(); iter != Values.end(); ++iter)
        { /* remove values */
         delete iter->second;
        } /* remove values */

     Values.clear();
     
     for(XMLValue::const_iterator iter = node->Values.begin(); iter != node->Values.end(); ++iter)
        { /* copy values */
         AddValue(iter->first, iter->second->GetValue(), iter->second->GetLine());
        } /* copy values */
    } // XMLNode::CopyValues

/****************************************************************************
*                            XMLNode::CopyChildren
* Inputs:
*       XMLNode * node: Node whose children are to be copied
* Result: void
*       
* Effect: 
*       Copies the children to the current node
****************************************************************************/

void XMLNode::CopyChildren(const XMLNode * node)
    {
     ASSERT(Children.empty());
     if(!Children.empty())
        { /* delete children */
         ASSERT(FALSE); // TODO: implement this
         return;        // TODO: for now, we return.  The tree is damaged
        } /* delete children */

     if(node->Children.empty())
        return;

     for(ChildNodes::const_iterator iter = node->Children.begin(); iter != node->Children.end(); ++iter)
        { /* copy children */
         XMLNode * newChild = Copy(*iter);
         Children.push_back(newChild);
        } /* copy children */
     
    } // XMLNode::CopyChildren

/****************************************************************************
*                                    XMLNode::Copy
* Inputs:
*       XMLNode * node: Node to copy
* Result: XMLNode *
*       Deep copy of the tree
* Effect: 
*       Makes a complete copy of the tree
****************************************************************************/

/* static */ XMLNode * XMLNode::Copy(const XMLNode * node)
    {
     if(node == NULL)
        return NULL;

     //****************************************************************
     // XMLNode
     //****************************************************************
     const XMLTag * tag = dynamic_cast<const XMLTag *>(node);
     if(tag != NULL)
        { /* XMLTag */
         XMLTag * result = new XMLTag(tag->GetTag(), tag->GetName(), tag->GetLine());
         result->CopyValues(tag);
         result->CopyChildren(tag);
         return reinterpret_cast<XMLNode*>(result);
        } /* XMLTag */
     //****************************************************************
     // XMLComment
     //****************************************************************
     const XMLComment * comment = dynamic_cast<const XMLComment *>(node);
     if(comment != NULL)
        { /* XMOComment */
         XMLComment * result = new XMLComment(comment->GetComment(), comment->GetLine());
         return reinterpret_cast<XMLNode*>(result);
        } /* XMOComment */
     //****************************************************************
     // XMLHeader
     //****************************************************************
     const XMLHeader * header = dynamic_cast<const XMLHeader*>(node);
     if(header != NULL)
        { /* header */
         XMLHeader * result = new XMLHeader(header->GetLine());
         result->CopyValues(header);
         result->CopyChildren(header);
         return reinterpret_cast<XMLNode*>(result);
        } /* header */

     //****************************************************************
     // XMLRoot
     //****************************************************************
     const XMLRoot * root = dynamic_cast<const XMLRoot*>(node);
     if(root != NULL)
        { /* XMLRoot */
         XMLRoot * result = new XMLRoot();
         result->CopyValues(root);
         result->CopyChildren(root);
         return reinterpret_cast<XMLNode*>(result);
        } /* XMLRoot */
     //****************************************************************
     // Unknown node
     //****************************************************************
     ASSERT(FALSE); // UNKNOWN node type
     return NULL;
    } // XMLNode::Copy

/****************************************************************************
*                               XMLNode::Remove
* Inputs:
*       XMLNode * child: Child node to remove from the child array
* Result: BOOL
*       TRUE if the node existed and was removed
*       FALSE if the node as not found
* Effect: 
*       Removes the node from the child array.  The node is deleted
****************************************************************************/

BOOL XMLNode::Remove(XMLNode * child)
    {
     for(ChildNodes::iterator iter = Children.begin(); iter != Children.end(); ++iter)
        { /* scan nodes */
         if((*iter) == child)
            { /* found it */
             XMLNode * node = *iter;
             Children.erase(iter);
             delete node;
             return TRUE;
            } /* found it */
        } /* scan nodes */
     return FALSE;
    } // XMLNode::Remove

/****************************************************************************
*                         XMLNode::RemoveAllChildren
* Result: void
*       
* Effect: 
*       Removes each child node and deltes it
****************************************************************************/

void XMLNode::RemoveAllChildren()
    {
     for(ChildNodes::iterator iter = Children.begin(); iter != Children.end(); ++iter)
        { /* delete each */
         delete *iter;
        } /* delete each */
     Children.clear();
    } // XMLNode::RemoveAllChildren

/****************************************************************************
*                              XMLNode::Replace
* Inputs:
*       XMLNode * OldNode:
*       XMLNode * NewNode:
* Result: BOOL
*       TRUE if the OldNode was replaced with the NewNode
*       FALSE if no replacement occurred
* Effect: 
*       Does a recursive tree-walk starting at 'this' and finds an instance
*       of OldNode, and replaces it with NewNode
*       If OldNode is replaced, it is freed, along with all of its child nodes
* Notes:
*       OldNode != this
****************************************************************************/

BOOL XMLNode::Replace(XMLNode * OldNode, XMLNode * NewNode)
    {
     ASSERT(OldNode != this);
     if(OldNode == this)
        return FALSE;
     for(size_t i = 0; i < Children.size(); ++i)
        { /* scan nodes */
         if(Children.at(i) == OldNode)
            { /* found it */
             Children.at(i) = NewNode;
             delete OldNode;
             return TRUE;
            } /* found it */
         else
         if(Children.at(i)->Replace(OldNode, NewNode))
            return TRUE;
        } /* scan nodes */
     return FALSE;
    } // XMLNode::Replace
