#pragma once

class XMLNode;

#ifdef XMLEXCEPTION_0B5F600D_7BF1_471e_B1AD_B3BE2A12445B
#define EXPORTED __declspec(dllexport)
#else
#define EXPORTED 
#endif
/****************************************************************************
*                                CXMLException
****************************************************************************/

class EXPORTED CXMLException : public CException
   {
    public:
       CXMLException(int ln);
       virtual ~CXMLException();
    public:
       int GetLine() { return line; }
       virtual CString GetString();
    protected:
       int line;
   };

/****************************************************************************
*                                CXMLValueException
****************************************************************************/
class EXPORTED CXMLDuplicateValueException : public CXMLException {
    public:
       CXMLDuplicateValueException(const CString & Old, int oldln, const CString & nm, const CString & val, int ln = 0);
       virtual ~CXMLDuplicateValueException();
       virtual CString GetString();
    protected:
       CString OldValue;
       int OldLine;
       CString Name;
       CString NewValue;
};

/****************************************************************************
*                                CXMLSyntaxException
****************************************************************************/
class CXMLSyntaxException : public CXMLException {
    public:
       CXMLSyntaxException(int ln, const CString & source, int offset);
       virtual ~CXMLSyntaxException();
       virtual CString GetString();
       CString GetText() { return SourceText; }
    protected:
       CString Found;
       CString SourceText;
};

/****************************************************************************
*                     class CXMLUnterminatedCommentException
****************************************************************************/
class CXMLUnterminatedCommentException : public CXMLSyntaxException {
    public:
       CXMLUnterminatedCommentException(int line, const CString & source, int offset) : CXMLSyntaxException(line, source, offset) { }
       virtual CString GetString();
};

/****************************************************************************
*                        class CXMLExpectedTokenException
****************************************************************************/
class CXMLExpectedTokenException : public CXMLSyntaxException {
    public:
       CXMLExpectedTokenException(const CString & exp, const CString & lexeme, int ln, const CString & source, int offset) : CXMLSyntaxException(ln, source, offset) {
          Expected = exp;
          Found = lexeme;
       }
       virtual CString GetString();
protected:
       CString Expected;
};


/****************************************************************************
*                       class CXMLExpectedNameException
****************************************************************************/
class  CXMLExpectedNameException : public CXMLSyntaxException {
    public:
       CXMLExpectedNameException(const CString & lexeme, int ln, const CString & source, int offset)
             : CXMLSyntaxException(ln, source, offset) { Found = lexeme; }
       virtual CString GetString();
       
};

/****************************************************************************
*                         CXMLExpectedValueException
****************************************************************************/

class CXMLExpectedValueException : public CXMLSyntaxException
    {
    public:
       CXMLExpectedValueException(const CString & lexeme, int ln, const CString & source, int offset)
             : CXMLSyntaxException(ln, source, offset) { Found = lexeme; }
       virtual CString GetString();
    }; // CXMLExpectedValueException

/****************************************************************************
*                         CXMLUnmatchedCloseException
****************************************************************************/

class CXMLUnmatchedCloseException : public CXMLSyntaxException {
    public:
       CXMLUnmatchedCloseException(const CString & otag, const CString & oname, int oline, const CString & etag, const CString & ename, int eline, const CString & source, int offset) :
             CXMLSyntaxException(eline, source, offset) {
          OpenTag = otag;
          OpenName = oname;
          OpenLine = oline;
          CloseTag = etag;
          CloseName = ename;
          CloseLine = eline;
       }
       virtual CString GetString();
    protected:
       CString OpenTag;
       CString OpenName;
       int OpenLine;
       CString CloseTag;
       CString CloseName;
       int CloseLine;
}; // CXMLUnmatchedCloseException
